@@ -67,8 +67,6 @@
 #include "sna_render_inline.h"
 #include "fb/fbpict.h"
 
-#include <mipict.h>
-
 #define FALLBACK 0
 #define NO_GLYPH_CACHE 0
 #define NO_GLYPHS_TO_DST 0
@@ -1142,6 +1140,11 @@ glyphs_via_mask(struct sna *sna,
 		if (mask_image == NULL)
 			goto err_pixmap;
 
+		if (sigtrap_get()) {
+			pixman_image_unref(mask_image);
+			goto err_pixmap;
+		}
+
 		memset(pixmap->devPrivate.ptr, 0, pixmap->devKind*height);
 #if HAS_PIXMAN_GLYPHS
 		if (sna->render.glyph_cache) {
@@ -1278,6 +1281,8 @@ next_image:
 		} while (--nlist);
 		pixman_image_unref(mask_image);
 
+		sigtrap_put();
+
 		mask = CreatePicture(0, &pixmap->drawable,
 				     format, CPComponentAlpha,
 				     &component_alpha, serverClient, &error);
@@ -2040,6 +2045,11 @@ glyphs_via_image(struct sna *sna,
 	if (mask_image == NULL)
 		goto err_pixmap;
 
+	if (sigtrap_get()) {
+		pixman_image_unref(mask_image);
+		goto err_pixmap;
+	}
+
 	memset(pixmap->devPrivate.ptr, 0, pixmap->devKind*height);
 #if HAS_PIXMAN_GLYPHS
 	if (sna->render.glyph_cache) {
@@ -2175,6 +2185,7 @@ next_image:
 			list++;
 		} while (--nlist);
 	pixman_image_unref(mask_image);
+	sigtrap_put();
 
 	component_alpha = NeedsComponent(format->format);
 
